/*
 * The contents of this file are subject to the MonetDB Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://monetdb.cwi.nl/Legal/MonetDBLicense-1.1.html
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the MonetDB Database System.
 *
 * The Initial Developer of the Original Code is CWI.
 * Portions created by CWI are Copyright (C) 1997-July 2008 CWI.
 * Copyright August 2008-2010 MonetDB B.V.
 * All Rights Reserved.
 */

#include "sql_config.h"
#include "mal_builder.h"
#include "mal_debugger.h"
#include "opt_prelude.h"
#include "sql_mvc.h"
#include "sql_optimizer.h"
#include "sql_scenario.h"
#include "sql_gencode.h"
#include "bpm/bpm_storage.h"
#include "replication.h"

#define TOSMALL 10

str 
FXoptimizer(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	InstrPtr *ns, p;
	int v;
	int i, k, newssize;
	lng clk = GDKusec();

#ifdef _SQL_OPTIMIZER_DEBUG
	stream_printf(GDKout, "First call to FXoptimizer\n");
	printFunction(GDKout, mb, 0, LIST_MAL_ALL);
#endif
	(void) cntxt;
	(void)stk;
	if( mb->errors) {
		/* remove the FXoptimizer request */
		if (pci)
			removeInstruction(mb, pci);
		return MAL_SUCCEED;
	}
	if( !factoryHasFreeSpace())
		return MAL_SUCCEED;



	if (mb->stop <= TOSMALL){
		for(i=0; i< mb->stop; i++){
			p= getInstrPtr(mb,0);
			if (getFunctionId(p)== bindRef && 
				getModuleId(p)== sqlRef)  break;
		}
		if(i == mb->stop)
			return MAL_SUCCEED;
	}


	newssize = mb->ssize * 2;
	ns = (InstrPtr *) GDKmalloc(sizeof(InstrPtr) * newssize);
	k = 1;
	ns[0] = getInstrPtr(mb, 0);	/* its signature */
	ns[0]->token = FACTORYsymbol;
 	setVarType(mb, getArg(ns[0],0), TYPE_bit);
	setVarUDFtype(mb,getArg(ns[0],0));
	for (i = 0; i < mb->stop; i++) {
		p = getInstrPtr(mb, i);
		if ( (getFunctionId(p) == bindRef || 
			  getFunctionId(p) == bindidxRef ||
			  getFunctionId(p) == binddbatRef ) && 
			getModuleId(p) == sqlRef) {
			ns[k++] = p;
		}
	}


	p = newInstruction(mb, ASSIGNsymbol);
	v = newTmpVariable(mb, TYPE_bit);
	p->barrier = BARRIERsymbol;
	p->argv[0] = v;
	pushBit(mb,p,TRUE);
	ns[k++] = p;

	for (i = 1; i < mb->stop - 1; i++) {
		p = getInstrPtr(mb, i);
		if ( !(getModuleId(p) == sqlRef &&
		     (getFunctionId(p) == bindRef || 
		      getFunctionId(p) == binddbatRef ||
		      getFunctionId(p) == bindidxRef)) ){
		} 
			ns[k++] = p;
	}


	p = newInstruction(mb,ASSIGNsymbol);
	p->barrier = YIELDsymbol;
	p->argv[0] = v;
	ns[k++] = p;
	p = newInstruction(mb,ASSIGNsymbol);
	p->barrier = REDOsymbol;
	p->argv[0] = v;
	ns[k++] = p;
	p = newInstruction(mb,ASSIGNsymbol);
	p->barrier = EXITsymbol;
	p->argv[0] = v;
	ns[k++] = p;
	ns[k++] = getInstrPtr(mb, i);

	mb->stop = k;
	mb->ssize = newssize;
	GDKfree(mb->stmt);
	mb->stmt = ns;

	optimizerCheck(cntxt,mb,"sql.factorize",1,GDKusec()-clk, OPT_CHECK_ALL);


#ifdef _SQL_OPTIMIZER_DEBUG
	printFunction(GDKout, mb, 0, LIST_MAL_STMT);
#endif
	return MAL_SUCCEED;
}



static void
SQLgetStatistics(Client cntxt, mvc *m, MalBlkPtr mb)
{
	InstrPtr *old = NULL;
	int oldtop, i, actions = 0, size = 0, j = 0;
	lng clk = GDKusec();

	old = mb->stmt;
	oldtop= mb->stop;
	size = (mb->stop *1.2 < mb->ssize)? mb->ssize: (int) (mb->stop *1.2);
	mb->stmt = (InstrPtr *) GDKzalloc(size  * sizeof(InstrPtr));
	mb->ssize = size ;
	mb->stop = 0;

	for(i=0; i<oldtop; i++){
		InstrPtr p = old[i];
		char *f = getFunctionId(p);

		if( getModuleId(p) == sqlRef && 
		    (f == bindRef || f == bindidxRef || f == binddbatRef ) ){
			ValRecord vr;
			InstrPtr q;
			char *sname = getVarConstant(mb, getArg(p,2)).val.sval;
			char *tname = getVarConstant(mb, getArg(p,3)).val.sval;
			char *cname = NULL;
			int not_null = 0;
			wrd rows = 1; /* default to cope with delta bats */
			int mode = 0;
			int part = 0;
			int k = getArg(p,0);
			sql_bpm *bpm = NULL;
			sql_schema *s = mvc_bind_schema(m, sname);

			if (!s || strcmp(s->base.name, dt_schema) == 0) {
				pushInstruction(mb,p);
				continue;
			}
			if (f == binddbatRef) {
				mode = getVarConstant(mb, getArg(p,4)).val.ival;
			} else {
				cname = getVarConstant(mb, getArg(p,4)).val.sval;
				mode = getVarConstant(mb, getArg(p,5)).val.ival;
			}


			if (s && f == bindidxRef && cname) {
				size_t cnt;
				sql_idx *i = mvc_bind_idx(m, s, cname);

				if (active_store_type == store_bpm && i->data) {
					bpm = i->data;
					part = bpm->nr;
				}
				cnt = store_funcs.count_idx(i);
				assert(cnt <= (size_t) GDK_oid_max);
				rows = (wrd) cnt;
			} else if (s && f == bindRef && cname) {
				size_t cnt;
				sql_table *t = mvc_bind_table(m, s, tname);
				sql_column *c = mvc_bind_column(m, t, cname);

				if (c) {
					not_null = !c->null;

					if (active_store_type == store_bpm && c->data) {
						bpm = c->data;
						part = bpm->nr;
					}
					cnt = store_funcs.count_col(c);
					assert(cnt <= (size_t) GDK_oid_max);
					rows = (wrd) cnt;
				}
			}	
			if (rows > 1 && mode != RD_INS)
				varSetProp(mb, k, rowsProp, op_eq, VALset(&vr, TYPE_wrd, &rows));
			if (not_null)
				varSetProp(mb, k, notnilProp, op_eq, NULL);
			if (0 && active_store_type == store_bpm && part && 
			    f != binddbatRef) {
				q = newInstruction(mb, ASSIGNsymbol);
				setModuleId(q, constraintsRef);
				setFunctionId(q, putName("parts", 5));
		        	setDestVar(q, newTmpVariable(mb, TYPE_void));
				for(j=p->retc; j<p->argc; j++) 
					q = pushArgument(mb, q, getArg(p,j));
				pushInt(mb, q, part);
				pushInstruction(mb,q);
			}
			if (part <= 1) {
				int lowprop = hlbProp, highprop = hubProp;
				/* rows == cnt has been checked above to be <= GDK_oid_max */
				oid low = 0, high = low + (oid)rows;
				pushInstruction(mb, p);

				if (mode == RD_INS) {
					if (f != binddbatRef)
						low = high;
					high += BPM_SPLIT;
				}
				if (f == binddbatRef) {
					lowprop = tlbProp;
					highprop = tubProp;
				}
				varSetProp(mb, getArg(p,0), lowprop, op_gte, VALset(&vr, TYPE_oid, &low));
				varSetProp(mb, getArg(p,0), highprop, op_lt, VALset(&vr, TYPE_oid, &high));
			} else {
				int i, v = getArg(p,0);
				int tpe = getVarType(mb, v);
				int top = mb->stop;
				oid low = 0, high = 0;

				for(i=0; i<part; i++){
					high += BPM_SPLIT;
					q = newInstruction(mb, ASSIGNsymbol);
					setModuleId(q, matRef);
					setFunctionId(q, f);
					getArg(q,0) = newTmpVariable(mb, tpe);
					for(j=p->retc; j<p->argc; j++) 
						q = pushArgument(mb, q, getArg(p,j));
					q = pushInt(mb, q, i);
					pushInstruction(mb,q);

					varSetProp(mb, getArg(q,0), hlbProp, op_gte, VALset(&vr, TYPE_oid, &low));
					varSetProp(mb, getArg(q,0), hubProp, op_lt, VALset(&vr, TYPE_oid, &high));

					low = high;
				}
				q = newInstruction(mb, ASSIGNsymbol);
				setModuleId(q,matRef);
				setFunctionId(q,newRef);
				getArg(q,0)= v;
				for(; top<mb->stop; top++) 
					q= pushArgument(mb, q, getArg(getInstrPtr(mb,top), 0));
				pushInstruction(mb,q);
			}
	
			if (not_null || part>1)
				actions++;
		} else {
			pushInstruction(mb,p);
		}
	}
	GDKfree(old);
	optimizerCheck(cntxt,mb,"optimizer.SQLgetstatistics",actions,GDKusec()-clk,0);
}


str minimalPlan= "inline,remap,deadcode,multiplex";

static str optimizers[256];	/* the optimizer pipeline */
static str optimizerpipe;		/* reference to last pipeline string */
static char *previouspipe = 0;
	
str
SQLvalidatePipeline(void){
	int mitosis= FALSE, deadcode= FALSE, mergetable= FALSE, replication= FALSE, multiplex=FALSE, garbage=FALSE;
	int i;

	mal_set_lock(sql_contextLock,"SQL optimizer");
	if (optimizers[0] &&  strcmp(optimizers[0],"inline") ) {
		mal_unset_lock(sql_contextLock,"SQL optimizer");
		throw(SQL,"optimizer"," 'inline' should be the first\n");
	}
	
	/* deadcode should be used */
	for ( i=0; optimizers[i]; i++)
		if (strcmp(optimizers[i],"deadcode") == 0)
			deadcode= TRUE;
		else
		if (strcmp(optimizers[i],"mitosis") == 0)
			mitosis= TRUE;
		else
		if (strcmp(optimizers[i],"mergetable") == 0)
			mergetable= TRUE;
		else
		if (strcmp(optimizers[i],"replication") == 0)
			replication= TRUE;
		else
		if (strcmp(optimizers[i],"multiplex") == 0)
			multiplex= TRUE;
		else
		if (strcmp(optimizers[i],"garbageCollector") == 0 && optimizers[i+1] == 0)
			garbage= TRUE;

#ifdef WIN32
		else
		if (strcmp(optimizers[i],"octopus") == 0){
			mal_unset_lock(sql_contextLock,"SQL optimizer");
			throw(SQL,"optimizer"," 'octopus' needs Linux\n");
		}
#endif
	if (optimizers[0] && mitosis == TRUE && mergetable == FALSE) {
		mal_unset_lock(sql_contextLock,"SQL optimizer");
		throw(SQL,"optimizer"," 'mitosis' needs 'mergetable'\n");
	}

	if (optimizers[0] && multiplex == 0){
		mal_unset_lock(sql_contextLock,"SQL optimizer");
		throw(SQL,"optimizer"," 'multiplex' should be used\n");
	}
	if (optimizers[0] && deadcode == FALSE ){
		mal_unset_lock(sql_contextLock,"SQL optimizer");
		throw(SQL,"optimizer"," 'deadcode' should be used at least once\n");
	}
	if (optimizers[0] && garbage == FALSE ){
		mal_unset_lock(sql_contextLock,"SQL optimizer");
		throw(SQL,"optimizer"," 'garbageCollector' should be used as the last one\n");
	}


	if ( MASTERroleEnabled() && replication == FALSE){
		mal_unset_lock(sql_contextLock,"SQL optimizer");
		throw(SQL,"optimizer"," 'replication' should be part of the pipeline\n");
	}

	mal_unset_lock(sql_contextLock,"SQL optimizer");
	return MAL_SUCCEED;
}



str
setOptimizers(str optimizer)
{
	int top=0;
	char *base, *nxt, *nme, *pipe="notdefined";
	
	/* do nothing if the pipe line is already set */
	if ( optimizerpipe && optimizer && strcmp(optimizerpipe,optimizer) == 0 )
		return optimizerpipe;

	/* catch  minimal pipes */
	if (optimizer == NULL || *optimizer == 0 )
		optimizer= minimalPlan;
	else
	/* optimizers can be temporarily turned on/off */
	if(strcmp(optimizer,"off")==0){
		if( previouspipe == NULL)
			previouspipe = optimizerpipe;
		else return optimizerpipe;
		optimizer= minimalPlan;
	} else
	if (strcmp(optimizer,"on")==0){
		if (previouspipe)
			optimizer= previouspipe;
		else return optimizerpipe;
		previouspipe = 0;
	} else {
		/* the optimizer may be an environment alias */
		pipe = GDKgetenv(optimizer);
		if ( pipe )
			optimizer = pipe;
	}

	optimizerpipe= optimizer;
	base = optimizer = GDKstrdup(optimizer);

	/* An optimizer pipe is a comma separated list of names */
	while (optimizer && *optimizer ) {
		nxt = strchr(optimizer,',');
		if (nxt){
			*nxt = 0;
			nxt++;
		}
		if ((nme = putName(optimizer,strlen(optimizer))) == 0)
			showException(SQL,"optimizer"," '%s' pipeline does not exist\n",optimizer);
		else if (top<256)
			optimizers[top++] = nme;
		else
			break;
		optimizer = nxt;
	} 
	GDKfree(base);
	if (top == 256){
		showException(SQL,"optimizer","Too many optimizer steps, use default pipe instead\n");
		setOptimizers("default_pipe");
		return "default_pipe";
	} 
	optimizers[top] = 0;
	if (top <= 1  && pipe == NULL){
		showException(SQL,"optimizer","Optimizer '%s' pipeline does not exist, use default pipe instead\n", optimizerpipe?optimizerpipe:"undefined");
		setOptimizers("default_pipe");
		return "default_pipe";
	}
	return optimizerpipe;
}

static void
addOptimizers(Client c,MalBlkPtr mb)
{
	int i;

	/* add the optimizers to the query plan */
	for (i = 0; optimizers[i]; i++) {
		InstrPtr p = newFcnCall(mb, "optimizer", optimizers[i]);
		typeChecker(c->nspace, mb, p, TRUE);
	}
}

void
addQueryToCache(Client c)
{
	backend *be = ((backend *) c->state[MAL_SCENARIO_PARSER]);
	str optimizer;
	MalBlkPtr mb;
	ValRecord src;
	mvc *m;

	insertSymbol(c->nspace, c->curprg);
	trimMalBlk(c->curprg->def);
	c->blkmode = 0;
	mb = c->curprg->def;
	chkProgram(c->nspace, mb);
	m = ((mvc *) c->state[MAL_SCENARIO_OPTIMIZE]);
#ifdef _SQL_OPTIMIZER_DEBUG
	stream_printf(GDKout, "ADD QUERY TO CACHE\n");
	printFunction(GDKout,mb,0,LIST_MAL_ALL);
#endif


	if (mb->errors) {
		mvc *m;
		m = ((mvc *) c->state[MAL_SCENARIO_OPTIMIZE]);
		showErrors(c);

		if (c->listing)
			printFunction(c->fdout, mb,0, c->listing);
		if ( m->debug )
			runMALDebugger(c,c->curprg);
		return;
	}
	optimizer = stack_get_string(be->mvc, "optimizer");
	stack_set_var(be->mvc, "optimizer", VALset(&src, TYPE_str, GDKstrdup(setOptimizers(optimizer))));
	addOptimizers(c,mb);
	SQLgetStatistics(c,(mvc *) c->state[MAL_SCENARIO_OPTIMIZE],mb);
	if ( m->emod & mod_debug )
		addtoMalBlkHistory(mb,"getStatistics");
	optimizeMALBlock(c,mb);

	/* time to execute the optimizers */
	if( c->debug)
		optimizerCheck(c,mb,"sql.baseline",-1,0, OPT_CHECK_ALL);
#ifdef _SQL_OPTIMIZER_DEBUG
	stream_printf(GDKout, "ADD optimized QUERY TO CACHE\n");
	printFunction(GDKout,mb,0,LIST_MAL_ALL);
#endif
}



str
SQLoptimizer(Client c)
{
	(void) c;
#ifdef _SQL_OPTIMIZER_DEBUG
	stream_printf(GDKout, "SQLoptimizer\n");
	printFunction(c->fdout, c->curprg->def,0, LIST_MAL_STMT);
	stream_printf(GDKout, "done\n");
#endif
	return MAL_SUCCEED;
}



