/*
 * The contents of this file are subject to the MonetDB Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://monetdb.cwi.nl/Legal/MonetDBLicense-1.1.html
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the MonetDB Database System.
 *
 * The Initial Developer of the Original Code is CWI.
 * Portions created by CWI are Copyright (C) 1997-July 2008 CWI.
 * Copyright August 2008-2010 MonetDB B.V.
 * All Rights Reserved.
 */


#include "sql_config.h"
#include "sql_result.h"
#include <str.h>
#include <tablet.h>
#include <mtime.h>
#include <bat/res_table.h>




int
dec_tostr(void *extra, char **Buf, int *len, int type, ptr a)
{
	/* support dec map to bte, sht, int and lng */
	if (type == TYPE_bte) {
		
	char buf[32];
	bte v = *(bte*)a;
	int scale = (int)(ptrdiff_t)extra, cur = 31, neg = (v<0)?1:0, i, done = 0;
	int l;

	if (v == bte_nil) {
		if (*len < 5){
			if (*Buf)
				GDKfree(*Buf);
			*len = 5;
			*Buf = GDKmalloc(*len);
		}
		strcpy(*Buf, "NULL");
		return 4;
	}
		
	if (v<0)
		v = -v;

	buf[cur--] = 0;
	if (scale){
		for (i=0; i<scale; i++) {
			buf[cur--] = (char) (v%10 + '0');
			v /= 10;
		}
		buf[cur--] = '.';
	}
	while (v) {
		buf[cur--] = (char ) (v%10 + '0');
		v /= 10;
		done = 1;
	}
	if (!done)
		buf[cur--] = '0';
	if (neg)
		buf[cur--] = '-';
	l = (32-cur-1);
	if (*len < l){
		if (*Buf)
			GDKfree(*Buf);
		*len = l+1;
		*Buf = GDKmalloc(*len);
	}
	strcpy(*Buf, buf+cur+1);
	return l-1;



	} else if (type == TYPE_sht) {
		
	char buf[32];
	sht v = *(sht*)a;
	int scale = (int)(ptrdiff_t)extra, cur = 31, neg = (v<0)?1:0, i, done = 0;
	int l;

	if (v == sht_nil) {
		if (*len < 5){
			if (*Buf)
				GDKfree(*Buf);
			*len = 5;
			*Buf = GDKmalloc(*len);
		}
		strcpy(*Buf, "NULL");
		return 4;
	}
		
	if (v<0)
		v = -v;

	buf[cur--] = 0;
	if (scale){
		for (i=0; i<scale; i++) {
			buf[cur--] = (char) (v%10 + '0');
			v /= 10;
		}
		buf[cur--] = '.';
	}
	while (v) {
		buf[cur--] = (char ) (v%10 + '0');
		v /= 10;
		done = 1;
	}
	if (!done)
		buf[cur--] = '0';
	if (neg)
		buf[cur--] = '-';
	l = (32-cur-1);
	if (*len < l){
		if (*Buf)
			GDKfree(*Buf);
		*len = l+1;
		*Buf = GDKmalloc(*len);
	}
	strcpy(*Buf, buf+cur+1);
	return l-1;



	} else if (type == TYPE_int) {
		
	char buf[32];
	int v = *(int*)a;
	int scale = (int)(ptrdiff_t)extra, cur = 31, neg = (v<0)?1:0, i, done = 0;
	int l;

	if (v == int_nil) {
		if (*len < 5){
			if (*Buf)
				GDKfree(*Buf);
			*len = 5;
			*Buf = GDKmalloc(*len);
		}
		strcpy(*Buf, "NULL");
		return 4;
	}
		
	if (v<0)
		v = -v;

	buf[cur--] = 0;
	if (scale){
		for (i=0; i<scale; i++) {
			buf[cur--] = (char) (v%10 + '0');
			v /= 10;
		}
		buf[cur--] = '.';
	}
	while (v) {
		buf[cur--] = (char ) (v%10 + '0');
		v /= 10;
		done = 1;
	}
	if (!done)
		buf[cur--] = '0';
	if (neg)
		buf[cur--] = '-';
	l = (32-cur-1);
	if (*len < l){
		if (*Buf)
			GDKfree(*Buf);
		*len = l+1;
		*Buf = GDKmalloc(*len);
	}
	strcpy(*Buf, buf+cur+1);
	return l-1;



	} else if (type == TYPE_lng) {
		
	char buf[32];
	lng v = *(lng*)a;
	int scale = (int)(ptrdiff_t)extra, cur = 31, neg = (v<0)?1:0, i, done = 0;
	int l;

	if (v == lng_nil) {
		if (*len < 5){
			if (*Buf)
				GDKfree(*Buf);
			*len = 5;
			*Buf = GDKmalloc(*len);
		}
		strcpy(*Buf, "NULL");
		return 4;
	}
		
	if (v<0)
		v = -v;

	buf[cur--] = 0;
	if (scale){
		for (i=0; i<scale; i++) {
			buf[cur--] = (char) (v%10 + '0');
			v /= 10;
		}
		buf[cur--] = '.';
	}
	while (v) {
		buf[cur--] = (char ) (v%10 + '0');
		v /= 10;
		done = 1;
	}
	if (!done)
		buf[cur--] = '0';
	if (neg)
		buf[cur--] = '-';
	l = (32-cur-1);
	if (*len < l){
		if (*Buf)
			GDKfree(*Buf);
		*len = l+1;
		*Buf = GDKmalloc(*len);
	}
	strcpy(*Buf, buf+cur+1);
	return l-1;



	} else {
		GDKerror("Decimal cannot be mapped to %s\n", ATOMname(type));
	}
	return 0;
}

struct time_res {
	int fraction;
	int has_tz;
	lng timezone;
};

int
sql_time_tostr(void *TS_RES, char **buf, int *len, int type, ptr A)
{
	struct time_res *ts_res = TS_RES;
	int i, len1, big = 128;
	char buf1[128], *s1 = buf1, *s;
	lng val = 0, timezone = ts_res->timezone * 60 * 1000;
	daytime tmp, *a = A;
	daytime mtime = 24 * 60 * 60 * 1000;

	(void) type;
	if (ts_res->has_tz) 
		val = *a + timezone;
	else
		val = *a;
	if (val < 0)
		val = mtime + val;
	if (val > mtime)
		val = val - mtime;
	tmp = (daytime) val;

	len1 = daytime_tostr(&s1, &big, &tmp);

	/* fixup the fraction, default is 3 */
	len1 += (ts_res->fraction-3);
	if (ts_res->fraction == 0) 
		len1 --;

	if (*len < len1 + 8) {
		if (*buf)
			GDKfree(*buf);
		*buf = (str) GDKmalloc(*len = len1 + 8);
	}
	s = *buf;
	strcpy(s, buf1);
	s += len1;
	s[0] = 0;
	/* extra zero's for usec's */
	for (i=3; i<ts_res->fraction; i++) 
		s[-i+2] = '0';

	if (ts_res->has_tz) {
		*s++ = (ts_res->timezone >= 0) ? '+' : '-';
		sprintf(s, "%02d:%02d", ABS(ts_res->timezone) / 60, ABS(ts_res->timezone) % 60);
		s += 5;
	}
	return (int) (s - *buf);
}

int
sql_timestamp_tostr(void *TS_RES, char **buf, int *len, int type, ptr A)
{
	struct time_res *ts_res = TS_RES;
	int i, len1, len2, big = 128;
	char buf1[128], buf2[128], *s, *s1 = buf1, *s2 = buf2;
	timestamp tmp, *a = A;
	lng timezone = ts_res->timezone * 60 * 1000;

	(void) type;
	if (ts_res->has_tz) {
		MTIMEtimestamp_add(&tmp, a, &timezone);
		len1 = date_tostr(&s1, &big, &tmp.days);
		len2 = daytime_tostr(&s2, &big, &tmp.msecs);
	} else {
		len1 = date_tostr(&s1, &big, &a->days);
		len2 = daytime_tostr(&s2, &big, &a->msecs);
	}

	/* fixup the fraction, default is 3 */
	len2 += (ts_res->fraction-3);
	if (ts_res->fraction == 0) 
		len2 --;

	if (*len < len1 + len2 + 8) {
		if (*buf)
			GDKfree(*buf);
		*buf = (str) GDKmalloc(*len = len1 + len2 + 8);
	}
	s = *buf;
	strcpy(s, buf1);
	s += len1;
	*s++ = ' ';
	strcpy(s, buf2);
	s += len2;
	s[0] = 0;
	/* extra zero's for usec's */
	for (i=3; i<ts_res->fraction; i++) 
		s[-i+2] = '0';

	if (ts_res->has_tz) {
		*s++ = (ts_res->timezone >= 0) ? '+' : '-';
		sprintf(s, "%02d:%02d", ABS(ts_res->timezone) / 60, ABS(ts_res->timezone) % 60);
		s += 5;
	}
	return (int) (s - *buf);
}

static int
bat_max_strlength(BAT *b)
{
	BUN p, q;
	int l = 0;
	int max = 0;
	BATiter bi = bat_iterator(b);

	BATloop(b, p, q) {
		str v = (str) BUNtail(bi, p);
		strLength(&l, v);

		if (l == int_nil) 
			l = 0;
		if (l > max)
			max = l;
	}
	return max;
}




static size_t
bat_max_btelength(BAT *b)
{
	BUN p, q;
	lng max = 0;
	lng min = 0;
	size_t ret = 0;
	BATiter bi = bat_iterator(b);

	BATloop(b, p, q) {
		lng m = 0;
		bte l = *((bte *)BUNtail(bi, p));

		if (l != bte_nil)
			m = l;
		if (m > max) max = m;
		if (m < min) min = m;
	}

	if ((-min * 10) > max) 
		max = (-min * 10);
	while (max /= 10) ret++;
		ret++;
	return ret;
}




static size_t
bat_max_shtlength(BAT *b)
{
	BUN p, q;
	lng max = 0;
	lng min = 0;
	size_t ret = 0;
	BATiter bi = bat_iterator(b);

	BATloop(b, p, q) {
		lng m = 0;
		sht l = *((sht *)BUNtail(bi, p));

		if (l != sht_nil)
			m = l;
		if (m > max) max = m;
		if (m < min) min = m;
	}

	if ((-min * 10) > max) 
		max = (-min * 10);
	while (max /= 10) ret++;
		ret++;
	return ret;
}




static size_t
bat_max_intlength(BAT *b)
{
	BUN p, q;
	lng max = 0;
	lng min = 0;
	size_t ret = 0;
	BATiter bi = bat_iterator(b);

	BATloop(b, p, q) {
		lng m = 0;
		int l = *((int *)BUNtail(bi, p));

		if (l != int_nil)
			m = l;
		if (m > max) max = m;
		if (m < min) min = m;
	}

	if ((-min * 10) > max) 
		max = (-min * 10);
	while (max /= 10) ret++;
		ret++;
	return ret;
}




static size_t
bat_max_lnglength(BAT *b)
{
	BUN p, q;
	lng max = 0;
	lng min = 0;
	size_t ret = 0;
	BATiter bi = bat_iterator(b);

	BATloop(b, p, q) {
		lng m = 0;
		lng l = *((lng *)BUNtail(bi, p));

		if (l != lng_nil)
			m = l;
		if (m > max) max = m;
		if (m < min) min = m;
	}

	if ((-min * 10) > max) 
		max = (-min * 10);
	while (max /= 10) ret++;
		ret++;
	return ret;
}







ptr *
dec_frstr(Column *c, int type, char *s, char *e, char quote)
{
	/* support dec map to bte, sht, int and lng */
	(void)e;
	(void)quote;
	if (s == e) {
		return NULL;
	} else if (type == TYPE_bte) {
		
	sql_column *col = c->extra;
	sql_subtype *t = &col->type;

	unsigned int i, neg = 0;
	bte *r;
	bte res = 0;
	if (*s == '-'){
		neg = 1;
		s++;
	} else if (*s == '+'){
		neg = 0;
		s++;
	}
	for (i = 0; i < (t->digits-t->scale) && *s != '.' && *s; i++, s++) {
		if (!*s || *s < '0' || *s > '9') 
			return NULL;
		res *= 10;
		res += (*s-'0');
	}
	if (!*s && t->scale) {
		for( i = 0; i < t->scale; i++) {
			res *= 10;
		}
	}
	if (*s) {
		if (*s != '.') 
			return NULL;
		s++;
		for( i = 0; *s && i < t->scale; i++, s++) {
			if (*s < '0' || *s > '9') 
				return NULL;
			res *= 10;
			res += (*s-'0');
		}
		for( ; i < t->scale; i++) {
			res *= 10;
		}
	}
	if (*s) 
		return NULL;
	r = c->data;
	if (!r)
		r = (bte*)GDKmalloc(sizeof(bte));
	c->data = r;
	if (neg)
		*r = -res;
	else
		*r = res;
	return (ptr*)r;



	} else if (type == TYPE_sht) {
		
	sql_column *col = c->extra;
	sql_subtype *t = &col->type;

	unsigned int i, neg = 0;
	sht *r;
	sht res = 0;
	if (*s == '-'){
		neg = 1;
		s++;
	} else if (*s == '+'){
		neg = 0;
		s++;
	}
	for (i = 0; i < (t->digits-t->scale) && *s != '.' && *s; i++, s++) {
		if (!*s || *s < '0' || *s > '9') 
			return NULL;
		res *= 10;
		res += (*s-'0');
	}
	if (!*s && t->scale) {
		for( i = 0; i < t->scale; i++) {
			res *= 10;
		}
	}
	if (*s) {
		if (*s != '.') 
			return NULL;
		s++;
		for( i = 0; *s && i < t->scale; i++, s++) {
			if (*s < '0' || *s > '9') 
				return NULL;
			res *= 10;
			res += (*s-'0');
		}
		for( ; i < t->scale; i++) {
			res *= 10;
		}
	}
	if (*s) 
		return NULL;
	r = c->data;
	if (!r)
		r = (sht*)GDKmalloc(sizeof(sht));
	c->data = r;
	if (neg)
		*r = -res;
	else
		*r = res;
	return (ptr*)r;



	} else if (type == TYPE_int) {
		
	sql_column *col = c->extra;
	sql_subtype *t = &col->type;

	unsigned int i, neg = 0;
	int *r;
	int res = 0;
	if (*s == '-'){
		neg = 1;
		s++;
	} else if (*s == '+'){
		neg = 0;
		s++;
	}
	for (i = 0; i < (t->digits-t->scale) && *s != '.' && *s; i++, s++) {
		if (!*s || *s < '0' || *s > '9') 
			return NULL;
		res *= 10;
		res += (*s-'0');
	}
	if (!*s && t->scale) {
		for( i = 0; i < t->scale; i++) {
			res *= 10;
		}
	}
	if (*s) {
		if (*s != '.') 
			return NULL;
		s++;
		for( i = 0; *s && i < t->scale; i++, s++) {
			if (*s < '0' || *s > '9') 
				return NULL;
			res *= 10;
			res += (*s-'0');
		}
		for( ; i < t->scale; i++) {
			res *= 10;
		}
	}
	if (*s) 
		return NULL;
	r = c->data;
	if (!r)
		r = (int*)GDKmalloc(sizeof(int));
	c->data = r;
	if (neg)
		*r = -res;
	else
		*r = res;
	return (ptr*)r;



	} else if (type == TYPE_lng) {
		
	sql_column *col = c->extra;
	sql_subtype *t = &col->type;

	unsigned int i, neg = 0;
	lng *r;
	lng res = 0;
	if (*s == '-'){
		neg = 1;
		s++;
	} else if (*s == '+'){
		neg = 0;
		s++;
	}
	for (i = 0; i < (t->digits-t->scale) && *s != '.' && *s; i++, s++) {
		if (!*s || *s < '0' || *s > '9') 
			return NULL;
		res *= 10;
		res += (*s-'0');
	}
	if (!*s && t->scale) {
		for( i = 0; i < t->scale; i++) {
			res *= 10;
		}
	}
	if (*s) {
		if (*s != '.') 
			return NULL;
		s++;
		for( i = 0; *s && i < t->scale; i++, s++) {
			if (*s < '0' || *s > '9') 
				return NULL;
			res *= 10;
			res += (*s-'0');
		}
		for( ; i < t->scale; i++) {
			res *= 10;
		}
	}
	if (*s) 
		return NULL;
	r = c->data;
	if (!r)
		r = (lng*)GDKmalloc(sizeof(lng));
	c->data = r;
	if (neg)
		*r = -res;
	else
		*r = res;
	return (ptr*)r;



	}
	return NULL;
}

ptr *
_ASCIIadt_frStr(Column *c, int type, char *s, char *e, char quote)
{
	int len;
	(void)quote;
	if (type == TYPE_str) {
		sql_column *col = (sql_column*)c->extra;

		if (col->type.digits > 0 && (unsigned)(e-s) > col->type.digits){
			str v = (str)s;
			strLength(&len, v);

			if (len == int_nil) 
				len = 4;
			if ((unsigned int) len > col->type.digits) 
				return NULL;
		}
		return TABLETstrFrStr(c, s, e);
	}
	 
	len = (*BATatoms[type].atomFromStr)(s, &c->len, (ptr) &c->data);
	if (len == 0 || len != e-s){
		/* decimals can be converted to integers when *.000 */
		if ( s[len++] == '.')
		switch( type ){
		case TYPE_bte: case TYPE_int: case TYPE_lng: case TYPE_sht:
			while (s[len]=='0') len++;
			if ( s[len] == 0)
				return c->data;
		}
		return NULL;
	}
	return c->data;
}


static int
_ASCIIadt_toStr(void *extra, char **buf, int *len, int type, ptr a)
{
	if (type == TYPE_str) {
		Column *c = extra;
		char *dst, *src = a;
		int l = (c->quote)?escapedStrlen(src):(int)strlen(src), l2 = 0;

		if (l + 3 > *len) {
			GDKfree(*buf);
			*len = 2 * l + 3;
			*buf = GDKmalloc(*len);
		}
		dst = *buf;
		if (c->quote) {
			dst[0] = c->quote;
			l2 = 1;
			l = escapedStr(dst+l2, src);
		} else {
			strncpy(dst + l2, src, l);
		}
		if (l2) {
			dst[l + l2] = c->quote;
			l2++;
		}
		dst[l + l2] = 0;
		return l + l2;
	} else {
		return (*BATatoms[type].atomToStr) (buf, len, a);
	}
}


BAT *
mvc_import_table(Client cntxt, mvc *m, bstream *bs, char *sname, char *tname, char *sep, char *rsep, char *ssep, char *ns, lng sz, lng offset)
{
	int i = 0;
	sql_schema *s = mvc_bind_schema(m, sname);
	sql_table *t = mvc_bind_table(m, s, tname);
	node *n;
	BAT *bats = NULL;
	Tablet as;
	Column *fmt;

	if (!t) {
		sql_error(m, 500, "table %s not found", tname);
		m->type = -1;
		return NULL;
	}
	if (!bs) {
		sql_error(m, 500, "no stream (pointer) provided");
		m->type = -1;
		return NULL;
	}
	if (stream_errnr(bs->s)) {
		sql_error(m, 500, "stream not open %d", stream_errnr(bs->s));
		m->type = -1;
		return NULL;
	}
	if (offset < 0 || offset > (lng) BUN_MAX) {
		sql_error(m, 500, "offset out of range");
		m->type = -1;
		return NULL;
	}

	if (offset > 0)
		offset--;
	if (t->columns.set) {
		stream *out = m->scanner.ws;

		memset((char*) &as, 0, sizeof(as));
		as.nr_attrs = list_length(t->columns.set);
		as.nr = (sz<1)?BUN_NONE:(BUN)sz;
		as.offset = (BUN) offset;
		as.error = NULL;
		as.tryall = 0;
		as.complaints= NULL;
		fmt = as.format = (Column *) GDKmalloc(sizeof(Column) * (as.nr_attrs + 1));
		if (!isa_block_stream(bs->s))
			 out = NULL;

		for (n = t->columns.set->h, i = 0; n; n = n->next, i++) {
			sql_column *col = n->data;

			fmt[i].name = col->base.name;
			fmt[i].sep = (n->next) ? _strdup(sep) : _strdup(rsep);
			fmt[i].seplen = _strlen(fmt[i].sep);
			fmt[i].type = col->type.type->base.name;
			fmt[i].adt = ATOMindex(fmt[i].type);
			fmt[i].tostr = &_ASCIIadt_toStr;
			fmt[i].frstr = &_ASCIIadt_frStr;
			fmt[i].extra = col;
			fmt[i].len = fmt[i].nillen = 
				ATOMlen(fmt[i].adt, ATOMnilptr(fmt[i].adt));
			fmt[i].data = GDKmalloc(fmt[i].len);
			fmt[i].c[0] = NULL;
			fmt[i].ws = !(has_whitespace(fmt[i].sep));
			fmt[i].quote = ssep?ssep[0]:0;
			fmt[i].nullstr = _strdup(ns);
			fmt[i].null_length = strlen(ns);
			fmt[i].nildata = GDKmalloc(fmt[i].nillen);
			memcpy(fmt[i].nildata, ATOMnilptr(fmt[i].adt), fmt[i].nillen);
			if (col->type.type->eclass == EC_DEC) {
				fmt[i].tostr = &dec_tostr;
				fmt[i].frstr = &dec_frstr;
			}
			fmt[i].raw = NULL;
			fmt[i].rawfile = NULL;
			fmt[i].batfile = NULL;
			fmt[i].size = ATOMsize(fmt[i].adt);
		}
		if (TABLETcreate_bats(&as, (BUN) (sz < 0 ? 1000 : sz)) >= 0) {
			if (SQLload_file(cntxt, &as, bs, out, sep, rsep, ssep?ssep[0]:0, offset, sz) != BUN_NONE && !as.error)
				bats = TABLETcollect_bats(&as);
		}
		if (as.error) 
			sql_error(m, 500, "%s", as.error);
		TABLETdestroy_format(&as);
	}
	return bats;
}




static int
mvc_export_warning(stream *s, str w) {
	str tmp = NULL;
	while (w != NULL && *w != '\0') {
		if ((tmp = strchr(w, (int)'\n')) != NULL)
			*tmp++ = '\0';
		if (stream_printf(s, "#%s", w) < 0)
			return(-1);
		w = tmp;
	}
	return(1);
}

int
mvc_export_prepare(mvc *c, stream *out, cq *q, str w)
{
	node *n;
	int len = c->params ? list_length(c->params) : 0;

	if (!out)
		return 0;

	/* write header, query type: Q_PREPARE */
	if (stream_printf(out,
		"&5 %d %d 3 %d\n" /* TODO: add type here: r(esult) or u(pdate) */
		"%% .prepare,	.prepare,	.prepare # table_name\n"
		"%% type,	digits,	scale # name\n"
		"%% varchar,	int,	int # type\n"
		"%% 0,	0,	0 # length\n",
		q->id, len, len) < 0) {
		return -1;
	}

	if (c->params) {
		int i;

		q->paramlen = len;
		q->params = SA_NEW_ARRAY(q->sa, sql_subtype, len);
		for (n = c->params->h, i=0; n; n = n->next, i++) {
			sql_arg *a = n->data;
			sql_subtype *t = &a->type;

			if (t) {
				if (stream_printf(out,
					"[ \"%s\",	%d,	%d	]\n",
					t->type->sqlname,
					t->digits,
					t->scale
				) < 0) {
					return -1;
				}
				/* add to the query cache parameters */
				q->params[i] = *t;
			} else {
				return -1;
			}
		}
	}
	if (mvc_export_warning(out, w) != 1)
		return -1;
	return 0;
}





int
mvc_send_bte(stream *s, bte cnt){
	char buf[50], *b;
	int neg = cnt <0;
	if(neg) cnt = -cnt;
	b= buf+49;
	do{
		*b--= (char) ('0'+ (cnt % 10));
		cnt /=10;
	} while(cnt>0);
	if( neg) 
		*b = '-'; 
	else b++;
	return stream_write(s, b, 50- (b-buf),1)==1;
}




int
mvc_send_sht(stream *s, sht cnt){
	char buf[50], *b;
	int neg = cnt <0;
	if(neg) cnt = -cnt;
	b= buf+49;
	do{
		*b--= (char) ('0'+ (cnt % 10));
		cnt /=10;
	} while(cnt>0);
	if( neg) 
		*b = '-'; 
	else b++;
	return stream_write(s, b, 50- (b-buf),1)==1;
}




int
mvc_send_int(stream *s, int cnt){
	char buf[50], *b;
	int neg = cnt <0;
	if(neg) cnt = -cnt;
	b= buf+49;
	do{
		*b--= (char) ('0'+ (cnt % 10));
		cnt /=10;
	} while(cnt>0);
	if( neg) 
		*b = '-'; 
	else b++;
	return stream_write(s, b, 50- (b-buf),1)==1;
}




int
mvc_send_lng(stream *s, lng cnt){
	char buf[50], *b;
	int neg = cnt <0;
	if(neg) cnt = -cnt;
	b= buf+49;
	do{
		*b--= (char) ('0'+ (cnt % 10));
		cnt /=10;
	} while(cnt>0);
	if( neg) 
		*b = '-'; 
	else b++;
	return stream_write(s, b, 50- (b-buf),1)==1;
}




int
convert2str( mvc *m, int eclass, int d, int sc, int has_tz, ptr p, int mtype, char **buf, int len)
{
	int l = 0;

	if (!p || ATOMcmp(mtype, ATOMnilptr(mtype), p) == 0) {
		(*buf)[0] = GDK_chr_min;
		(*buf)[1] = 0;
	} else if (eclass == EC_DEC) {
		l = dec_tostr((void *) (ptrdiff_t) sc, buf, &len, mtype, p);
	} else if (eclass == EC_TIME) {
		struct time_res *ts_res = alloca(sizeof(struct time_res));
		ts_res->has_tz = has_tz;
		ts_res->fraction = d?d-1:0;
		ts_res->timezone = m->timezone;
		l = sql_time_tostr( (void *) ts_res, buf, &len, mtype, p);

	} else if (eclass == EC_TIMESTAMP) {
		struct time_res *ts_res = alloca(sizeof(struct time_res));
		ts_res->has_tz = has_tz;
		ts_res->fraction = d?d-1:0;
		ts_res->timezone = m->timezone;
		l = sql_timestamp_tostr((void *) ts_res, buf, &len, mtype, p);
	} else if (eclass == EC_BIT) {
		bit b = *(bit*)p;
		if (b == bit_nil) {
			(*buf)[0] = 'N';
			(*buf)[1] = 'U';
			(*buf)[2] = 'L';
			(*buf)[3] = 'L';
			(*buf)[4] = 0;
		} else if (b) {
			(*buf)[0] = '1';
			(*buf)[1] = 0;
		} else {
			(*buf)[0] = '0';
			(*buf)[1] = 0;
		}
	} else {
		l = (*BATatoms[mtype].atomToStr) (buf, &len, p);
	}
	return l;
}

int
export_value( mvc *m, stream *s, int eclass, char *sqlname, int d, int sc, ptr p, int mtype, char **buf, int *len, str ns)
{
	int ok = 0;
	int l = 0;

	if (!p || ATOMcmp(mtype, ATOMnilptr(mtype), p) == 0) {
		size_t ll = strlen(ns);
		ok = (stream_write(s, ns, ll, 1) == 1);
	} else if (eclass == EC_DEC) {
		l = dec_tostr((void *) (ptrdiff_t) sc, buf, len,
				  mtype, p);
		ok = (stream_write(s, *buf, l, 1) == 1);
	} else if (eclass == EC_TIME) {
		struct time_res *ts_res = alloca(sizeof(struct time_res));
		ts_res->has_tz = (strcmp(sqlname, "timetz") == 0);
		ts_res->fraction = d?d-1:0;
		ts_res->timezone = m->timezone;
		l = sql_time_tostr( (void *) ts_res, buf, len, mtype, p);

		ok = (stream_write(s, *buf, l, 1) == 1);
	} else if (eclass == EC_TIMESTAMP) {
		struct time_res *ts_res = alloca(sizeof(struct time_res));
		ts_res->has_tz = (strcmp(sqlname, "timestamptz") == 0);
		ts_res->fraction = d?d-1:0;
		ts_res->timezone = m->timezone;
		l = sql_timestamp_tostr((void *) ts_res, buf, len, mtype, p);

		ok = (stream_write(s, *buf, l, 1) == 1);
	} else {
		switch(mtype) {
		case TYPE_bte:
			ok = mvc_send_bte(s, *(bte*)p);
			break;
		case TYPE_sht:
			ok = mvc_send_sht(s, *(sht*)p);
			break;
		case TYPE_int:
			ok = mvc_send_int(s, *(int*)p);
			break;
		case TYPE_lng:
			ok = mvc_send_lng(s, *(lng*)p);
			break;
		default: {
			l = (*BATatoms[mtype].atomToStr) (buf, len, p);
			ok = (stream_write(s, *buf, l, 1) == 1);
		}
		}
	}
	return ok;
}

static int
mvc_export_row(mvc *m, stream *s, res_table *t, str btag, str sep, str rsep, str ssep, str ns)
{
	size_t seplen = strlen(sep);
	size_t rseplen= strlen(rsep);
	char *buf = NULL;
	int len = 0;
	int i, ok = 1;

	if (!s)
		return 0;

	(void)ssep;
	if (btag[0])
		ok = (stream_write(s, btag, strlen(btag), 1) == 1);
	for (i = 0; i < t->nr_cols && ok; i++) {
		res_col *c = t->cols + i;

		if (i != 0) {
			ok = (stream_write(s, sep, seplen, 1) == 1);
			if (!ok)
				break;
		}
		ok = export_value(m, s, c->type.type->eclass, 
			c->type.type->sqlname, c->type.digits, c->type.scale, 
			c->p, c->mtype, &buf, &len, ns);
	}
	if (len)
		_DELETE(buf);
	if (ok)
		ok = (stream_write(s, rsep, rseplen, 1) == 1);
	m->results = res_tables_remove(m->results, t);
	return (ok) ? 0 : -1;
}

static int
mvc_export_table(mvc *m, stream *s, res_table *t, BAT *order, BUN offset, BUN nr, char *btag, char *sep, char *rsep, char *ssep, char *ns )
{
	Tablet as;
	Column *fmt;
	int i;

	if (!t)
		return -1;
	if (!s)
		return 0;

	as.nr_attrs = t->nr_cols + 1;	/* for the leader */
	as.nr = nr;
	as.offset = offset;
	fmt = as.format = (Column *) GDKmalloc(sizeof(Column) * (as.nr_attrs + 1));

	memset(fmt, 0, (as.nr_attrs + 1) * sizeof(Column));

	fmt[0].c[0] = NULL;
	fmt[0].sep = _strdup(btag);
	fmt[0].seplen = _strlen(fmt[0].sep);
	fmt[0].ws = 0;
	fmt[0].nullstr = NULL;

	for (i = 1; i <= t->nr_cols; i++) {
		res_col *c = t->cols + (i - 1);

		if (!c->b)
			break;

		fmt[i].c[0] = BATdescriptor(c->b);
		fmt[i].ci[0] = bat_iterator(fmt[i].c[0]);
		fmt[i].name = NULL;
		fmt[i].sep = ((i - 1) < (t->nr_cols - 1)) ? _strdup(sep) : _strdup(rsep);
		fmt[i].seplen = _strlen(fmt[i].sep);
		fmt[i].type = ATOMname(fmt[i].c[0]->ttype);
		fmt[i].adt = fmt[i].c[0]->ttype;
		fmt[i].tostr = &_ASCIIadt_toStr;
		fmt[i].frstr = &_ASCIIadt_frStr;
		fmt[i].extra = fmt+i;
		fmt[i].data = NULL;
		fmt[i].len = 0;
		fmt[i].nillen = 0;
		fmt[i].ws = 0;
		fmt[i].quote = ssep?ssep[0]:0;
		fmt[i].nullstr = _strdup(ns);
		if (c->type.type->eclass == EC_DEC) {
			fmt[i].tostr = &dec_tostr;
			fmt[i].frstr = &dec_frstr;
			fmt[i].extra = (void *) (ptrdiff_t) c->type.scale;
		} else if (c->type.type->eclass == EC_TIMESTAMP) {
			struct time_res *ts_res = alloca(sizeof(struct time_res));
			ts_res->has_tz = (strcmp(c->type.type->sqlname, "timestamptz") == 0); 
			ts_res->fraction = c->type.digits?c->type.digits-1:0;
			ts_res->timezone = m->timezone;

			fmt[i].tostr = &sql_timestamp_tostr;
			fmt[i].frstr = NULL;
			fmt[i].extra = ts_res;
		} else if (c->type.type->eclass == EC_TIME) {
			struct time_res *ts_res = alloca(sizeof(struct time_res));
			ts_res->has_tz = (strcmp(c->type.type->sqlname, "timetz") == 0); 
			ts_res->fraction = c->type.digits?c->type.digits-1:0;
			ts_res->timezone = m->timezone;

			fmt[i].tostr = &sql_time_tostr;
			fmt[i].frstr = NULL;
			fmt[i].extra = ts_res;
		} else {
			fmt[i].extra = fmt+i;
		}
	}
	if (i == t->nr_cols + 1) {
		TABLEToutput_file(&as, order, s);
	}
	TABLETdestroy_format(&as);
	if (stream_errnr(s))
		return -1;
	return 0;
}

static int 
export_length( stream *s, int mtype, int eclass, int digits, int scale, int tz, bat bid, ptr p) 
{
	int ok = 1;
	size_t count = 0;

	mtype = ATOMstorage(mtype);
	if (mtype == TYPE_str) {
		if (eclass == EC_CHAR) {
			ok = mvc_send_int(s, digits);
		} else {
			int l = 0;
			if (bid) {
				BAT *b = BATdescriptor(bid);

				if (b) {
					l = bat_max_strlength(b);
					BBPunfix(b->batCacheid);
				} else {
					assert(b);
					/* [Stefan.Manegold@cwi.nl]:
					 * Instead of an assert() or simply ignoring the problem,
					 * we could/should return an error code, but I don't know
					 * which it the correct/suitable error code -1|0|1 ?
					 *
					 return -1|0|1 ;
					 */
				}
			} else if (p) {
				str v = (str) p;

				strLength(&l, v);
				if (l == int_nil) 
					l = 0;
			}
			ok = mvc_send_int(s, l);
		}
	} else if (eclass == EC_NUM) {
		count = 0;
		if (bid) {
			BAT *b = BATdescriptor(bid);

			if (b) {
				if (mtype == TYPE_bte) {
					count = bat_max_btelength(b);
				} else if (mtype == TYPE_sht) {
					count = bat_max_shtlength(b);
				} else if (mtype == TYPE_int) {
					count = bat_max_intlength(b);
				} else { /* TYPE_lng */
					count = bat_max_lnglength(b);
				}
				BBPunfix(b->batCacheid);
			} else {
				assert(b);
				/* [Stefan.Manegold@cwi.nl]:
				 * Instead of an assert() or simply ignoring the problem,
				 * we could/should return an error code, but I don't know
				 * which it the correct/suitable error code -1|0|1 ?
				 *
				 return -1|0|1 ;
				 */
			}
		} else {
			if (p) {
				lng val = 0;
				if (mtype == TYPE_bte) {
					val = *((bte *)p);
				} else if (mtype == TYPE_sht) {
					val = *((sht *)p);
				} else if (mtype == TYPE_int) {
					val = *((int *)p);
				} else { /* TYPE_lng */
					val = *((lng *)p);
				}

				if (val < 0) count++;
				while (val /= 10) count++;
				count++;
			} else {
				count = 0;
			}
		}
		ok = mvc_send_lng(s, (lng) count);
	/* the following two could be done once by taking the
	   max value and calculating the number of digits from that
	   value, instead of the maximum values taken now, which
	   include the optional sign */
	} else if (eclass == EC_FLT) {
		/* [sign]+digit+period+[max 6 digits]+E+[sign]+[max 2 digits] */
		if (mtype == TYPE_flt) {
			ok = mvc_send_int(s, 13);
		/* [sign]+digit+period+[max 14 digits]+E+[sign]+[max 3 digits] */
		} else { /* TYPE_dbl */
			ok = mvc_send_int(s, 22);
		}
	} else if (eclass == EC_DEC) {
		count = 1 + digits;
		if (scale > 0)
			count += 1;
		ok = mvc_send_lng(s, (lng) count);
	} else if (eclass == EC_DATE) {
		ok = mvc_send_int(s, 10);
	} else if (eclass == EC_TIME) {
		count = 8;
		if (tz)	/* time zone */
			count += 6;	/* +03:30 */
		if (digits > 1)	/* fractional seconds precision (including dot) */
			count += digits;
		ok = mvc_send_lng(s, (lng) count);
	} else if (eclass == EC_TIMESTAMP) {
		count = 10 + 1 + 8;
		if (tz)	/* time zone */
			count += 6;		/* +03:30 */
		if (digits)	/* fractional seconds precision */
			count += digits;
		ok = mvc_send_lng(s, (lng) count);
	} else if (eclass == EC_BIT) {
		ok = mvc_send_int(s, 5);	/* max(strlen("true"), strlen("false")) */
	} else {
		ok = mvc_send_int(s, 0);
	}
	return ok;
}

 
int
mvc_export_value(mvc *m, stream *s, int qtype, str tn, str cn, str type, int d, int sc, int eclass, ptr p, int mtype, str w, str ns) 
{
	char *buf = NULL;
	int len = 0;
	int ok = 1;
	char *rsep = "\t]\n";

#ifdef NDEBUG
	(void) qtype; /* pacify compiler in case asserts are disabled */
#endif
	assert(qtype == Q_TABLE);

	if (stream_write(s, "&1 0 1 1 1\n", 11, 1) == 1 &&
	    /* fallback to default tuplecount (1) and id (0) */
	    /* TODO first header name then values */
	    stream_write(s, "% ", 2, 1) == 1 &&
	    stream_write(s, tn, strlen(tn), 1) == 1 &&
	    stream_write(s, " # table_name\n% ", 16, 1) == 1 &&
	    stream_write(s, cn, strlen(cn), 1) == 1 &&
	    stream_write(s, " # name\n% ", 10, 1) == 1 &&
	    stream_write(s, type, strlen(type), 1) == 1 &&
	    stream_write(s, " # type\n% ", 10, 1) == 1 &&
	    export_length(s, mtype, eclass, d, sc, has_tz(eclass,type), 0, p) &&
	    stream_write(s, " # length\n[ ", 12, 1) == 1 &&
	    export_value( m, s, eclass, type, d, sc, p, mtype, &buf, &len, ns)) 
		ok = (stream_write(s, rsep, strlen(rsep), 1) == 1);
	    
	if (buf)
		_DELETE(buf);

	if (ok)
		ok = mvc_export_warning(s, w);
	return ok;
}
int
mvc_export_operation(mvc *m, stream *s, str w)
{
	assert(m->type == Q_SCHEMA || m->type == Q_TRANS);
	if (m->type == Q_SCHEMA) {
		if (!s || stream_write(s, "&3\n", 3, 1) != 1)
			return -1;
	} else {
		if (m->session->auto_commit) {
			if (stream_write(s, "&4 t\n", 5, 1) != 1)
				return -1;
		} else {
			if (stream_write(s, "&4 f\n", 5, 1) != 1)
				return -1;
		}
	}

	if (mvc_export_warning(s, w) != 1)
		return -1;
	return 0;
}

int
mvc_export_affrows(mvc *m, stream *s, lng val, str w)
{
	/* if we don't have a stream, nothing can go wrong, so we return
	 * success.  This is especially vital for execution of internal SQL
	 * commands, since they don't get a stream to suppress their output.
	 * If we would fail on having no stream here, those internal commands
	 * fail too.
	 */
	if (!s) return 0;
	
	if (stream_write(s, "&2 ", 3, 1) != 1 || 
	   !mvc_send_lng(s, val) ||
	    stream_write(s, " ", 1, 1) != 1 ||
	   !mvc_send_lng(s, m->last_id) ||
   	    stream_write(s, "\n", 1, 1) != 1)
		return -1;
	if (mvc_export_warning(s, w) != 1)
		return -1;

	m->last_id = -1; /* reset after we exposed the value */

	return 0;
}

static int
export_error(BAT *order)
{
	if (order)
		BBPunfix(order->batCacheid);
	return -1;
}

int
mvc_export_head(mvc *m, stream *s, int res_id, int only_header)
{
	int i, res = 0;
	BUN count = 0;
	res_table *t = res_tables_find(m->results, res_id);
	BAT *order = NULL;

	if (!s || !t)
		return 0;

	/* query type: Q_TABLE */
	if (!(stream_write(s, "&1 ", 3, 1) == 1))
		return -1;

	/* id */
	if (!mvc_send_int(s, t->id) ||
			stream_write(s, " ", 1, 1) != 1)
		return -1;

	/* tuple count */
	if (only_header) {
		if (t->order) {
			order = BATdescriptor(t->order);
			if (!order)
				return -1;

			count = BATcount(order);
		} else
			count = 1;
	}
	if (!mvc_send_lng(s, (lng) count) ||
			stream_write(s, " ", 1, 1) != 1)
		return export_error(order);

	/* column count */
	if (!mvc_send_int(s, t->nr_cols) ||
			stream_write(s, " ", 1, 1) != 1)
		return export_error(order);

	/* row count, min(count, reply_size) */
	if (!mvc_send_int(s, (BUN) m->reply_size < count ? m->reply_size : (int) count)) 
		return export_error(order);

	if (stream_write(s, "\n% ", 3, 1) != 1)
		return export_error(order);
	for (i = 0; i < t->nr_cols; i++) {
		res_col *c = t->cols + i;
		size_t len = strlen(c->tn);

		if (len && stream_write(s, c->tn, len, 1) != 1)
			return export_error(order);
		if (i + 1 < t->nr_cols && stream_write(s, ",\t", 2, 1) != 1)
			return export_error(order);
	}
	if (stream_write(s, " # table_name\n% ", 16, 1) != 1)
		return export_error(order);

	for (i = 0; i < t->nr_cols; i++) {
		res_col *c = t->cols + i;

		if (stream_write(s, c->name, strlen(c->name), 1) != 1)
			return export_error(order);
		if (i + 1 < t->nr_cols && stream_write(s, ",\t", 2, 1) != 1)
			return export_error(order);
	}
	if (stream_write(s, " # name\n% ", 10, 1) != 1)
		return export_error(order);

	for (i = 0; i < t->nr_cols; i++) {
		res_col *c = t->cols + i;

		if (stream_write(s, c->type.type->sqlname,
					strlen(c->type.type->sqlname), 1) != 1)
			return export_error(order);
		if (i + 1 < t->nr_cols && stream_write(s, ",\t", 2, 1) != 1)
			return export_error(order);
	}
	if (stream_write(s, " # type\n% ", 10, 1) != 1)
		return export_error(order);

	for (i = 0; i < t->nr_cols; i++) {
		res_col *c = t->cols + i;
		int mtype = c->type.type->localtype;
		int eclass = c->type.type->eclass;

		if (!export_length(s, mtype, eclass, c->type.digits, c->type.scale, type_has_tz(&c->type), c->b, c->p))
			return export_error(order);
		if (i + 1 < t->nr_cols && stream_write(s, ",\t", 2, 1) != 1)
			return export_error(order);
	}
	if (stream_write(s, " # length\n", 10, 1) != 1)
		return export_error(order);

	if ( order)
		BBPunfix(order->batCacheid);
	return res;
}

static int
mvc_export_file(mvc *m, stream *s, res_table *t) 
{
	int res = 0;
	BUN count;
	BAT *order = NULL;

	if (m->scanner.ws == s) 
		/* need header */
		mvc_export_head(m, s, t->id, TRUE);

	if (!t->order) {
		res = mvc_export_row(m, s, t, "", t->tsep, t->rsep, t->ssep, t->ns);
	} else {
		order = BATdescriptor(t->order);
		if (!order)
			return -1;
		count = BATcount(order);

		res = mvc_export_table(m, s, t, order, 0, count, "", t->tsep, t->rsep, t->ssep, t->ns);
		BBPunfix(order->batCacheid);
		m->results = res_tables_remove(m->results, t);
	} 
	return res;
}

int
mvc_export_result(mvc *m, stream *s, int res_id)
{
	int clean = 0, res = 0;
	BUN count;
	res_table *t = res_tables_find(m->results, res_id);
	BAT *order = NULL;

	if (!s || !t)
		return 0;

	/* we shouldn't have anything else but Q_TABLE here */
	assert(t->query_type == Q_TABLE);
	if (t->tsep)
		return mvc_export_file(m, s, t);

	mvc_export_head(m, s, res_id, TRUE);

	if (!t->order)
		return mvc_export_row(m, s, t, "[ ", ",\t", "\t]\n", "\"", "NULL");
	order = BATdescriptor(t->order);
	if (!order)
		return -1;

	count = m->reply_size;
	if (count == 0 || count >= BATcount(order)) {
		count = BATcount(order);
		clean = 1;
	}
	res = mvc_export_table(m, s, t, order, 0, count, "[ ", ",\t", "\t]\n", "\"", "NULL");
	BBPunfix(order->batCacheid);
	if (clean)
		m->results = res_tables_remove(m->results, t);

	if (res > 0)
		res = mvc_export_warning(s, "");
	return res;
}

int
mvc_export_chunk(mvc *m, stream *s, int res_id, BUN offset, BUN nr)
{
	int res = 0;
	res_table *t = res_tables_find(m->results, res_id);
	BAT *order = NULL;
	BUN cnt;

	if (!s || !t)
		return 0;


	/* query type: Q_BLOCK */
	if (!(stream_write(s, "&6 ", 3, 1) == 1))
		return export_error(order);

	/* result id */
	if (!mvc_send_int(s, res_id) ||
			stream_write(s, " ", 1, 1) != 1)
		return export_error(order);

	/* column count */
	if (!mvc_send_int(s, t->nr_cols) ||
			stream_write(s, " ", 1, 1) != 1)
		return export_error(order);

	order = BATdescriptor(t->order);
	if (!order)
		return -1;
	cnt = nr;
	if (cnt == 0) 
		cnt = BATcount(order);
	if (offset >= BATcount(order))
		cnt = 0;
	if (offset + cnt > BATcount(order)) 
		cnt = BATcount(order) - offset;

	/* row count */
	if (!mvc_send_lng(s, (lng) cnt) || 
			stream_write(s, " ", 1, 1) != 1)
		return export_error(order);

	/* block offset */
	if (!mvc_send_lng(s, (lng) offset)) 
		return export_error(order);

	if (stream_write(s, "\n", 1, 1) != 1)
		return export_error(order);

	res = mvc_export_table(m, s, t, order, offset, cnt, "[ ", ",\t", "\t]\n", "\"", "NULL");
	BBPunfix(order->batCacheid);
	return res;
}


int
mvc_result_table(mvc *m, int nr_cols, int type, BAT *order)
{
	res_table *t = res_table_create(m->session->tr, m->result_id++, nr_cols, type, m->results, order);
	m->results = t;
	return t->id;
}

int
mvc_result_column(mvc *m, char *tn, char *name, char *typename, int digits, int scale, BAT *b)
{
	(void)res_col_create(m->session->tr, m->results, tn, name, typename, digits, scale, TYPE_bat, b);
	return 0;
}

int
mvc_result_value(mvc *m, char *tn, char *name, char *typename, int digits, int scale, ptr *p, int mtype)
{
	(void)res_col_create(m->session->tr, m->results, tn, name, typename, digits, scale, mtype, p);
	return 0;
}


